@@ -1,23 +1,23 @@
 # -*- coding: utf-8 -*-
+from __future__ import absolute_import, print_function, unicode_literals
 
 import os.path
-
-from django.template import Template, RequestContext
-from aldryn_apphooks_config.utils import get_app_instance
+from copy import deepcopy
 
 from cms import api
 from cms.apphook_pool import apphook_pool
-from cms.utils import get_cms_setting
+from cms.utils.conf import get_cms_setting
+from django.conf import settings
 from django.core.urlresolvers import reverse
 from django.http import SimpleCookie
+from django.template import RequestContext, Template
 from django.utils.encoding import force_text
 from django.utils.six import StringIO
-from django.conf import settings
 from djangocms_helper.base_test import BaseTestCase
 
-
+from ..utils import get_app_instance, get_apphook_configs, get_apphook_field_names
 from .utils.example.models import (
-    AnotherExampleConfig, ExampleConfig, Article, News, TranslatableArticle
+    AnotherExampleConfig, Article, ExampleConfig, News, NotApphookedModel, TranslatableArticle,
 )
 
 
@@ -28,6 +28,10 @@ class AppHookConfigTestCase(BaseTestCase):
         self.language = settings.LANGUAGES[0][0]
         self.root_page = api.create_page(
             'root page', self.template, self.language, published=True)
+        # This is needed in django CMS 3.5+ to keep the same tree across
+        # all django CMS versions
+        if hasattr(self.root_page, 'set_as_homepage'):
+            self.root_page.set_as_homepage()
 
         self.ns_app_1 = ExampleConfig.objects.create(namespace='app1')
         self.ns_app_1.app_data.config.property = 'app1_property'
@@ -37,6 +41,10 @@ class AppHookConfigTestCase(BaseTestCase):
         self.ns_app_2.app_data.config.property = 'app2_property'
         self.ns_app_2.app_data.config.published_default = True
         self.ns_app_2.save()
+        self.ns_app_3 = AnotherExampleConfig.objects.create(namespace='app3')
+        self.ns_app_3.app_data.config.property = 'app3_property'
+        self.ns_app_3.app_data.config.published_default = True
+        self.ns_app_3.save()
 
         self.page_1 = api.create_page(
             'page_1', self.template, self.language, published=True,
@@ -57,6 +65,7 @@ class AppHookConfigTestCase(BaseTestCase):
             for language, _ in settings.LANGUAGES[1:]:
                 api.create_title(language, page.get_slug(), page)
                 page.publish(language)
+        self.reload_urlconf()
 
     def test_configs(self):
         app = apphook_pool.get_apphook(self.page_1.application_urls)
@@ -74,7 +83,7 @@ class AppHookConfigTestCase(BaseTestCase):
         request = self.get_page_request(self.page_3, self.user)
         config = get_app_instance(request)
         # when config is requested on a non-config apphook, just return empty data
-        self.assertEqual((u'', None), config)
+        self.assertEqual(('', None), config)
 
     def test_no_page(self):
         request = self.request_factory.get('/en/sample/login/')
@@ -86,7 +95,7 @@ class AppHookConfigTestCase(BaseTestCase):
         # when config is requested on a non-CMS url, just return empty data
         with self.settings(ROOT_URLCONF='cms.test_utils.project.urls'):
             config = get_app_instance(request)
-            self.assertEqual((u'', None), config)
+            self.assertEqual(('', None), config)
 
     def test_config_str(self):
         app = apphook_pool.get_apphook(self.page_1.application_urls)
@@ -169,9 +178,9 @@ class AppHookConfigTestCase(BaseTestCase):
                             slug='news_2_app_1_config2',
                             section=self.ns_app_1,
                             config=ans_config_2)
-        msg = ("'{0}' has {1} relations to an ApphookConfig model."
-               " Please, specify which one to use in argument 'to'."
-               " Choices are: {2}".format('News', '2', 'section, config'))
+        msg = ('"{0}" has {1} relations to an ApphookConfig model.'
+               ' Please, specify which one to use in argument "to".'
+               ' Choices are: {2}'.format('News', '2', 'section, config'))
         self.assertRaisesMessage(
             ValueError, msg, News.objects.namespace, ans_config_1.namespace
         )
@@ -226,12 +235,14 @@ class AppHookConfigTestCase(BaseTestCase):
 
         # correct parameter passed by the request
         request = self.get_page_request(self.page_3, self.user)
+        request.GET = deepcopy(request.GET)
         request.GET['section'] = self.ns_app_1.pk
         retrieved = admin_instance.get_config_data(request, article, 'property')
         self.assertEqual(retrieved, self.ns_app_1.property)
 
         # correct parameter passed by the request - no existing object
         request = self.get_page_request(self.page_3, self.user)
+        request.GET = deepcopy(request.GET)
         request.GET['section'] = self.ns_app_1.pk
         retrieved = admin_instance.get_config_data(request, Article(), 'property')
         self.assertEqual(retrieved, self.ns_app_1.property)
@@ -288,6 +299,7 @@ class AppHookConfigTestCase(BaseTestCase):
 
         # object is set, normal form is used
         request = self.get_page_request(self.page_3, self.user)
+        request.GET = deepcopy(request.GET)
         request.GET['section'] = self.ns_app_1.pk
         form = admin_instance.get_form(request, article)
         self.assertEqual(list(form.base_fields.keys()), ['title', 'slug', 'section', 'published'])
@@ -295,6 +307,7 @@ class AppHookConfigTestCase(BaseTestCase):
 
         # no object is set, parameter passed through the request
         request = self.get_page_request(self.page_3, self.user)
+        request.GET = deepcopy(request.GET)
         request.GET['section'] = self.ns_app_1.pk
         form = admin_instance.get_form(request, None)
         self.assertEqual(list(form.base_fields.keys()), ['title', 'slug', 'section', 'published'])
@@ -311,6 +324,38 @@ class AppHookConfigTestCase(BaseTestCase):
         self.assertEqual(list(form.base_fields.keys()), ['title', 'slug', 'section', 'published'])
         self.assertEqual(form.base_fields['section'].initial, self.ns_app_1)
 
+    def test_apphook_admin(self):
+        from django.contrib import admin
+        admin.autodiscover()
+
+        admin_instance = admin.site._registry[ExampleConfig]
+        request = self.get_page_request(self.page_3, self.user)
+
+        # Testing Readonly field
+        self.assertEqual(
+            admin_instance.get_readonly_fields(request), ('type',)
+        )
+        self.assertEqual(
+            admin_instance.get_readonly_fields(request, self.ns_app_1), ('type', 'namespace')
+        )
+
+        # Testing admin output for sample app specific implementation
+        response = admin_instance.change_view(request, str(self.ns_app_1.pk))
+        try:
+            self.assertContains(
+                response,
+                '<div class="readonly">aldryn_apphooks_config.tests.utils.example.cms_appconfig.ExampleConfig</div>'
+            )
+            self.assertContains(response, '<div class="readonly">app1</div>')
+            self.assertContains(response, 'value="app1_property"')
+        except AssertionError:
+            self.assertContains(
+                response,
+                '<p>aldryn_apphooks_config.tests.utils.example.cms_appconfig.ExampleConfig</p>'
+            )
+            self.assertContains(response, '<p>app1</p>')
+            self.assertContains(response, 'name="config-property" type="text" value="app1_property"')
+
     def test_admin(self):
         from django.contrib import admin
         admin.autodiscover()
@@ -332,13 +377,22 @@ class AppHookConfigTestCase(BaseTestCase):
         response = admin_instance.add_view(request)
         self.assertContains(response, '$(this).apphook_reload_admin')
         self.assertContains(response, 'aldryn_apphooks_config')
-        self.assertContains(response, '<option value="1" selected="selected">%s</option>' % self.ns_app_1)
-        self.assertContains(response, '<input id="id_published"')
+        self.assertRegexpMatches(
+            force_text(response.content),
+            '(<option value="1" selected="selected">%s</option>|<option value="1" selected>%s</option>)' % (
+                self.ns_app_1, self.ns_app_1
+            )
+        )
+        self.assertContains(response, 'id="id_published"')
 
         self.ns_app_1.app_data.config.published_default = True
         self.ns_app_1.save()
         response = admin_instance.add_view(request)
-        self.assertContains(response, '<input checked="checked" id="id_published"')
+        response.render()
+        self.assertRegexpMatches(
+            force_text(response.content),
+            '(checked id="id_published"|id="id_published" checked|<input checked="checked" id="id_published")'
+        )
 
     def test_templatetag(self):
         article = Article.objects.create(title='news_1_app_1_config1',
@@ -364,4 +418,47 @@ class AppHookConfigTestCase(BaseTestCase):
         context = RequestContext(request, {'object': article, 'current_app': self.ns_app_2.namespace})
         template = Template('{% load apphooks_config_tags %}{% namespace_url "example_list" %}')
         response = template.render(context)
-        self.assertEqual(response, self.page_2.get_absolute_url())
\ No newline at end of file
+        self.assertEqual(response, self.page_2.get_absolute_url())
+
+    def test_apphook_field_name_discovery(self):
+        field_names = get_apphook_field_names(Article)
+        self.assertEqual(field_names, ['section'])
+
+        field_names = get_apphook_field_names(TranslatableArticle)
+        self.assertEqual(field_names, ['section'])
+
+        field_names = get_apphook_field_names(News)
+        self.assertEqual(set(field_names), set(['config', 'section']))
+
+        field_names = get_apphook_field_names(NotApphookedModel)
+        self.assertEqual(field_names, [])
+
+    def test_apphook_field_name_discovery_from_objects(self):
+        field_names = get_apphook_field_names(Article())
+        self.assertEqual(field_names, ['section'])
+
+        field_names = get_apphook_field_names(TranslatableArticle())
+        self.assertEqual(field_names, ['section'])
+
+        field_names = get_apphook_field_names(News())
+        self.assertEqual(set(field_names), set(['config', 'section']))
+
+        field_names = get_apphook_field_names(NotApphookedModel())
+        self.assertEqual(field_names, [])
+
+    def test_apphook_config_objects_discovery(self):
+        obj = Article(section=self.ns_app_1)
+        configs = get_apphook_configs(obj)
+        self.assertEqual(configs, [self.ns_app_1])
+
+        obj = TranslatableArticle(section=self.ns_app_1)
+        configs = get_apphook_configs(obj)
+        self.assertEqual(configs, [self.ns_app_1])
+
+        obj = News(section=self.ns_app_1, config=self.ns_app_3)
+        configs = get_apphook_configs(obj)
+        self.assertEqual(set(configs), set([self.ns_app_1, self.ns_app_3]))
+
+        obj = NotApphookedModel()
+        configs = get_apphook_configs(obj)
+        self.assertEqual(configs, [])
